!Defines a parameterization, computes likelihoods and defined proposal density
!Change this file to change these things
!The Cls are computed using routines defined in the CMB_Cls module

module ParamDef
  !module defines a parameterization, and works out power spectra
 use CMB_Cls
 use cmbtypes
 use cmbdata
 use Random
 use settings
 implicit none

 Type ParamSet
    real P(num_params)
    Type(ParamSetInfo) Info
 end Type

 Type ParamScale
    real PMin(num_params), PMax(num_params), PWidth(num_params)
 end Type ParamScale

 Type(ParamScale) Scales

 integer, dimension(:), allocatable :: params_used,fast_params_used
 integer num_params_used, num_fast_params_used
 
 logical :: has_propose_matrix = .false.
 real, dimension(:,:), allocatable :: propose_matrix
 real, dimension(:), allocatable :: propose_diag

contains


 
subroutine Initialize(Params)
        use IniFile
        implicit none
        type (ParamSet) Params
        integer i
        character(LEN=5000) InLine
        character(LEN=120) prop_mat
        real center, wid, mult, like
        logical restart
        real pmat(num_params,num_params)
        
        Ini_fail_on_not_found = .false.
        call CMB_Initialize(Params%Info)
        prop_mat = Ini_Read_String('propose_matrix')
        has_propose_matrix = prop_mat /= ''

        InLine = Ini_Read_String('continue_from')
  
        restart = InLine /= ''
   
        if (restart) then
           InLine = LastFileLine(InLine)
      
           read(InLine, *) mult, like, Params%P(1:num_params)
        end if
    
        num_params_used = 0
        num_fast_params_used = 0
        do i=1,num_params
   
           InLine = Ini_Read_String(numcat('param',i), .true.)
          
           read(InLine, *, err = 100) center, Scales%PMin(i), Scales%PMax(i), wid, Scales%PWidth(i)
           if (Scales%PMax(i) < Scales%PMin(i)) stop 'You have param Max < Min'
           if (Scales%PWidth(i) /= 0) then
               num_params_used = num_params_used + 1
              if (i > num_hard) num_fast_params_used = num_fast_params_used + 1

           end if
           if (.not. restart) then
           do
            if (wid < 0) then
              !This case we want half gaussian, width -wid
              !e.g. for positive definite parameters
              Params%P(i) = center - abs(Gaussian1())*wid
            else
              Params%P(i) = center + Gaussian1()*wid
            end if
            !Repeat until get acceptable values in range
            if (Params%P(i)>=  Scales%PMin(i) .and. Params%P(i) <= Scales%PMax(i)) exit
      
           end do
           end if
        end do

        if (Feedback > 0 ) write(*,'(" Varying ",1I2," parameters (",1I2," fast)")') &
               num_params_used,num_fast_params_used
        
        allocate(params_used(num_params_used))
        allocate(fast_params_used(num_fast_params_used))

        num_params_used = 0
        num_fast_params_used = 0
        do i=1,num_params
           if (Scales%PWidth(i) /= 0) then
              num_params_used = num_params_used + 1
              params_used(num_params_used) = i
              if (i > num_hard) then
                num_fast_params_used = num_fast_params_used + 1
                fast_params_used(num_fast_params_used) = i          
              end if
           end if
        end do
   

        if (has_propose_matrix) then
           
           call ReadMatrix(prop_mat,pmat, num_params, num_params)
           !If generated with constrained parameters, assume diagonal in those parameters
           do i=1,num_params
              if (pmat(i,i) ==0 .and. Scales%PWidth(i)/=0) pmat(i,i) = Scales%PWidth(i)**2
           !Enforce new constraints
           if (Scales%PWidth(i)==0) then
              pmat(i,:) = 0
              pmat(:,i) = 0
           end if
           end do
           
           allocate(propose_matrix(num_params_used, num_params_used))
           allocate(propose_diag(num_params_used))
           propose_matrix = pmat(params_used, params_used)
           call Diagonalize(propose_matrix,propose_diag, num_params_used)
           propose_diag = sqrt(max(0.,propose_diag))
           
        end if
   

        return
100     write(*,*) 'Error reading param details: '//trim(InLIne)
        stop

end subroutine Initialize

 
  subroutine WriteCMBParams(CMB,Theory,mult,like)     
     Type (CosmoTheory) Theory 
     real, intent(in) :: mult, like
     Type(ParamSet) P
     Type(CMBParams) CMB

     call CMBParamsToParams(CMB,P%P)
     
     P%Info%Theory = Theory
     call WriteParams(P, mult,like)

  end subroutine WriteCMBParams

   
  subroutine WriteIndepSample(P, like)
    Type(ParamSet) P
    real like
    Type(CMBParams) C

    if (indepfile_unit ==0) return

    call ParamsToCMBParams(P%P,C)

    call WriteModel(indepfile_unit, C,P%Info%Theory,like)
    if (flush_write) call FlushFile(indepfile_unit)

   end subroutine WriteIndepSample

end module ParamDef
